///////////////////////////////////////////////
//
// SolVBE
// 
// "universal" VESA VBE for windows dos boxes
//
// Copyright 2004 Jari Komppa 
// http://iki.fi/sol
//
///////////////////////////////////////////////
// License
///////////////////////////////////////////////
// 
// This software is provided 'as-is', without any express or implied
// warranty.  In no event will the authors be held liable for any damages
// arising from the use of this software.
// 
// Permission is granted to anyone to use this software for any purpose,
// including commercial applications, and to alter it and redistribute it
// freely, subject to the following restrictions:
// 
// 1. The origin of this software must not be misrepresented; you must not
//    claim that you wrote the original software. If you use this software
//    in a product, an acknowledgment in the product documentation would be
//    appreciated but is not required.
// 2. Altered source versions must be plainly marked as such, and must not be
//    misrepresented as being the original software.
// 3. This notice may not be removed or altered from any source distribution.
// 
// (ie. same as ZLIB license)
// 
///////////////////////////////////////////////
// Note
///////////////////////////////////////////////
//
// If you make any interesting additions / fixes, please send me patches,
// so that there's one "good" version of SolVBE.
//
// In order to build, SolVBE requires:
// - OpenWatcom 1.2 for DOS-side code
// - VC6 and Win2k DDK (other DDKs might also work)
//
// Some of the code might seem rather confusing. gRegs[AX] will alter the
// registers from host application (say, terra nova: strike force centauri),
// while setAX(n) alters the registers for the dos-side code of SolVBE.
//
///////////////////////////////////////////////
//
// This file:
// - VDD interface
// - Globals
//
///////////////////////////////////////////////

#include "SolVBE.h"



///////////////////////////////////////////////
// Globals
///////////////////////////////////////////////

                // 12345678901234567890123456789012345678901234567890123456789012345678901234567890
char gDLLInfo[] = "\tSolVBE VDD DLL Version 1.0b, March 8, 2004\n\tCopyright 2004 Jari Komppa\n";

HANDLE gDLLHandle;             // DLL handle
HWND   gWindowHandle;          // Display window
HWND   gConsoleWindow;         // Console handle (for keyboard events)
HDC    gWindowDC;              // Window device context
HGLRC  gOpenGLRC;              // OpenGL rendering context
GLuint gTextureHandle;         // OpenGL texture handle

HDC            gVideoMemoryDC;    // Video memory buffer handle
HBITMAP        gVideoMemoryOldBM; // Old bitmap / video memory
HBITMAP        gVideoMemoryBM;    // Current video memory bitmap
unsigned char *gVideoMemoryPtr;   // Pointer to the video memory

HDC     gFrameBufferDC;        // Framebuffer buffer handle
HBITMAP gFrameBufferOldBM;     // Old bitmap handle / fb
HBITMAP gFrameBufferBM;        // Framebuffer bitmap
int    *gFrameBufferPtr;       // Pointer to the framebuffer

int gLogmode = 0;              // log mode; don't log duplicate events
int gLogmodestart = 0;         // start tick for current log mode

volatile int gThreadAlive = 1; // Flag that keeps display update thread alive
volatile int gModeChanging = 0;// Flag that keeps SolVBE from shutting down when changing modes
int gLastVRC = 0;              // Last vertical retrace tick
volatile int gVRC = 0;         // Flag: are we inside vertical retrace?
int gStartTick = 0;            // Start tick for this SolVBE instance
CRITICAL_SECTION gThreadBusy;  // Rendering thread synchronization mutex
HANDLE gThreadHandle;          // Display update thread

int gVideoBufferWidth = 320;   // Current video buffer width
int gVideoBufferPitch = 320;   // Current pitch
int gVideoBufferHeight = 200;  // Current video buffer height
int gScreenWidth = 320;  // Current screen x resolution
int gScreenHeight = 200; // Current screen y resolution
int gWindowWidth = 320;  // Current window x resolution
int gWindowHeight = 200; // Current window y resolution
int gTextureWidth = 512; // Current texture x resolution
int gTextureHeight = 256;// Current texture y resolution

// Explanation on the above:
// video buffer - The contents of video memory. This is most likely
//                something like 320x6553. Games may change the display
//                offset for page flipping effects.
// pitch        - How many bytes we need to add to the pointer to get
//                to the next scanline. Depends on pixel formats.
// Screen       - The visible area of the above.
// Window       - The window resolution. May be anything, and image is
//                stretched to the resolution.
// Texture      - Textures usually have to be in powers-of-two, so for
//                a screen sized 320x200 we need a 512x256 texture.

int gBankOffset = 0;              // Current bank offset
int gDisplayXOffset = 0;          // Current display x-offset
int gDisplayYOffset = 0;          // Current display y-offset
int gSurfaceMode = 0;             // Current surface mode. Needed for setscanlinelength call.
int gLinearFiltering = 1;         // Flag: linear filtering
int gLockAspect = 1;              // Flag: aspect ratio locked
int gFullscreen = 0;              // Flag: fullscreen mode
int gPortsCaptured = 0;           // Flag: are VGA registers captured?

int gPalette[256];                // Current palette (4/8b modes)
unsigned int gPalIndex = 0;       // Current palette index (inp/outp)
unsigned int gPalWidth = 6;       // Current palette width (6 or 8 bit)
int gSeqReg04h = (2+4+8);         // Sequencer register 4h (bit 8 = chain-4)
int gSeqReg02h = (1);             // Sequencer register 2h (layer write mask)
int gCRTCReg17h = (128+64+32+8+4+2+1); // CRTC register 17h (bit 7 = byte addressing)
int gCRTCReg14h = (64 + 32);      // CRTC register 14h (bit 6 = dword addresses)
int gCRTCReg09h = (128 + 64 + 32);// CRTC register 9h (bit 8 = scan doubling)
int gGraphicsReg04h = 0;          // Graphics register 4h (read layer)

int gSeqAddress = 0;              // Active sequencer register
int gCRTCAddress = 0;             // Active CRTC address
int gGraphicsAddress = 0;         // Active graphics register

int gTweakedMode = 0;             // Flag: tweaked mode

int gCurrentMode = 3;             // Current video mode (3 = text)

int gMouseEventNeeded = 0;        // Bit flags on what kinds of things have happened since last event.
int gMouseCaptured = 0;           // Flag: is the mouse captured?
int gMouseButtons = 0;            // Currently depressed mouse buttons
int gMouseX = 0;                  // Current mouse x coordinate
int gMouseY = 0;                  // Current mouse y coordinate
int gMouseMinX = 0;               // Mouse min. x coordinate
int gMouseMaxX = 320;             // Mouse max. x coordinate
int gMouseMinY = 0;               // Mouse min. y coordinate
int gMouseMaxY = 200;             // Mouse max. y coordinate
int gMouseXSpeed = 8;             // Mouse x speed (mickeys per pixel; lower = faster)
int gMouseYSpeed = 8;             // Mouse y speed (mickeys per pixel; lower = faster)
int gMouseEventMask = 0;          // Mouse event call mask

int gMouseClickCount[2] = {0, 0}; // Current mouse click count
int gMouseClickX[2] = {0, 0};     // X Coordinate of last mouse click
int gMouseClickY[2] = {0, 0};     // Y Coordinate of last mouse click
int gMouseReleaseCount[2] = {0, 0};  // Current mouse release count
int gMouseReleaseX[2] = {0, 0};   // X Coordinate of last mouse release
int gMouseReleaseY[2] = {0, 0};   // Y Coordinate of last mouse release
int gMickeyX = 0;                 // Current X mickey count
int gMickeyY = 0;                 // Current Y mickey count

unsigned short *gRegs;            // Array of registers. (valid while processing interrupts)
int gRMDataSeg;                   // Real mode data segment
int gRMInfoOfs;                   // Real mode info offset (in data seg)
int gRMModeOfs;                   // Real mode mode list offset (in data seg)
int gRMCodeSeg;                   // Real mode code segment
int gRMBankSwitchOfs;             // Real mode bank switch function offset (in code seg)
int gRMMouseTrapOfs;              // Real mode far call calling function offset (in code seg)
unsigned char *gRMA000Ptr;        // Pointer to the real-mode A000 segment

__int64 gTickDivisor = 1;   // PerformanceCounter divisor for milliseconds
__int64 gTickStart = 0;     // PerformanceCounter start value
int gUseHighPerformanceTimer = 0;

///////////////////////////////////////////////
// Time
///////////////////////////////////////////////

int tick()
{
    if (gUseHighPerformanceTimer)
    {
        __int64 v;
        QueryPerformanceCounter((LARGE_INTEGER *)&v);
        v -= gTickStart;
        v /= gTickDivisor;
        return (int)v;
    }
    return GetTickCount();  
}

///////////////////////////////////////////////
// Debug functions
///////////////////////////////////////////////

void log(char * aFormat, ...)
{
    va_list marker;
    va_start(marker, aFormat);
#ifdef DEBUG_OUTPUT_LOG
    FILE * f = fopen("c:\\SolVBE.log","ab+");
    vfprintf(f, aFormat, marker);
    fclose(f);
#endif
}

void log_tick()
{
    int currenttick = tick();
    if (gLogmode != LOG_NORMAL)
    {
        gLogmode = LOG_NORMAL;
        log(" (%dms)\r\n", currenttick - gLogmodestart);
    }
    log("%8dms:",currenttick - gStartTick);
}


///////////////////////////////////////////////
// Finding the console
///////////////////////////////////////////////

void findConsoleWindow()
{
    char oldtitle[200];
    GetConsoleTitle(oldtitle,199);
    srand(GetTickCount());
    char title[81];
    int i;
    for (i = 0; i < 80; i++)    
        title[i] = (rand()%27)+'A';
    title[80] = 0;
    memcpy(title,"(SolVBE Console Detect) ", 24);                  
    SetConsoleTitle(title);
    Sleep(50); // make sure the title has changed
    gConsoleWindow = FindWindow(NULL,title);
    SetConsoleTitle(oldtitle);
    if (gConsoleWindow == NULL)
    {
        if (MessageBox(NULL,"Could not find the console window.\n"
                        "\n"
                        "Keyboard will not work.\n\n"
                        "Are you sure you want to continue?",
                        "SolVBE Error",MB_YESNO|MB_ICONWARNING) == IDNO)
        {
            VDDTerminateVDM();
        }
    }
}


///////////////////////////////////////////////
// Main functions
///////////////////////////////////////////////


/*
*
* VDDInitialize
*
* Arguments:
*     DllHandle - Identifies this VDD
*     Reason - Attach or Detach
*     Context - Not Used
*
* Return Value:
*     TRUE: SUCCESS
*     FALSE: FAILURE
*
*/

BOOL DllMain(IN PVOID aDllHandle,
             IN ULONG aReason,
             IN PCONTEXT aContext OPTIONAL)
{    
    int i;
    switch (aReason) 
    {       
    case DLL_PROCESS_ATTACH:
        {
            if (QueryPerformanceFrequency((LARGE_INTEGER *)&gTickDivisor))
            {
                gTickDivisor /= 1000;
                QueryPerformanceCounter((LARGE_INTEGER *)&gTickStart);
                gUseHighPerformanceTimer = 1;
            }


            gStartTick = tick();
            struct tm *newtime;
            long ltime;
            time( &ltime );

            // Obtain coordinated universal time: 
            newtime = gmtime( &ltime );
            
            // wipe log
#ifdef DEBUG_OUTPUT_LOG
            FILE * f = fopen("c:\\SolVBE.log","w");
            fclose(f);
#endif
            log_tick();
            log("==================================================\r\n");
            log_tick();
            log("SolVBE initialized at %s\r\n", asctime( newtime ));
        }            

        gDLLHandle = aDllHandle;
        gRMA000Ptr = (unsigned char *)VdmMapFlat(0xA000,0,VDM_V86);        
        if (gRMA000Ptr == NULL)
            MessageBox(NULL,"Unable to map a000 segment","SolVBE",MB_OK|MB_ICONERROR);
        
        InitializeCriticalSection(&gThreadBusy);

        gThreadHandle = CreateThread(
            NULL,
            0,
            threadproc,
            NULL,
            0,
            NULL);
        

        for (i = 0; i < 256; i++)
            gPalette[i] = i * 0xf47b4b35;
        break;
        
    case DLL_PROCESS_DETACH:
        gThreadAlive = 0;
        Sleep(500);
        DeleteCriticalSection(&gThreadBusy);
        DestroyWindow(gWindowHandle);
        VdmUnmapFlat(0xgRMA000Ptr,0,gRMA000Ptr,VDM_V86);

        releasePorts();
        {
            struct tm *newtime;
            long ltime;
            time( &ltime );

            // Obtain coordinated universal time:
            newtime = gmtime( &ltime );
            log_tick();
            log("SolVBE deinitialized at %s\r\n", asctime( newtime ));
            log_tick();
            log("--------------------------------------------------\r\n");
        }
        
        break;
        
    default:
        break;
    }
    
    return TRUE;
}


/*
*
* VDDInit
*
*/
extern "C" VOID SolVBEInit( VOID )
{
    
    // Called from dos side. Since we do all initialization
    // in dllmain, we can just set carry clear here to signal "ok".
    
    setCF( 0 );
        
    return;
}





extern "C" VOID SolVBECall( VOID )
{
    if (getBX() == 0x0074)
    {
        // mouse irq event
        domouseevent(); 
        return;
    }
    if (getBX() == 0x0100)
    {
        // configure call 1
        gRMDataSeg = getDS();
        gRMInfoOfs = getCX();
        gRMModeOfs = getDX();
        char *infoptr = (char *)VdmMapFlat(gRMDataSeg, gRMInfoOfs, VDM_V86);
        strcpy(infoptr, gDLLInfo);
        VdmUnmapFlat(gRMDataSeg, gRMInfoOfs, infoptr, VDM_V86);
        return;
    }
    if (getBX() == 0x0200)
    {
        // configure call 2
        findConsoleWindow(); // find the console window handle
        gRMCodeSeg = getCS();
        gRMBankSwitchOfs = getCX();        
        gRMMouseTrapOfs = getDX();
        return;
    }
    if (getBX() == 0x0033)
    {
        // Mouse interrupt
        gRegs = (unsigned short *)VdmMapFlat(getSS(),getBP(),VDM_V86);
        
        handle_int33();       

        VdmUnmapFlat(getSS(),getBP(),gRegs,VDM_V86);
        return;
    }
    if (getBX() == 0x0010)
    {
        // Video interrupt
        gRegs = (unsigned short *)VdmMapFlat(getSS(),getBP(),VDM_V86);
        
        handle_int10();        

        VdmUnmapFlat(getSS(),getBP(),gRegs,VDM_V86);
        return;
    }    
    char temp[1024];
    sprintf(temp,"Unknown call: BX=%04x\n(version mismatch?)",getBX());
    MessageBox(NULL,temp,"SolVBE Unknown Call",MB_OK|MB_ICONERROR);
    VDDTerminateVDM();
}
